import asyncio

from .backends.cameras.base import CameraBackend
from .backends.triggers.input.base import TriggerInput
from .config.app import CfgApp


class CameraApp:
    def __init__(self, camera: CameraBackend, trigger_input: TriggerInput):
        self.__config = CfgApp()

        self.__camera = camera
        self.__trigger_input = trigger_input

    async def setup(self):
        asyncio.create_task(self.__camera.run())
        # asyncio.create_task(self.__trigger.run())

    async def job_task(self):
        while True:
            try:
                job_uuid = await asyncio.wait_for(self.__trigger_input.receive_job_id(), timeout=0.5)
            except TimeoutError:
                # use wait_for with timeout since otherwise receive_job_id would block for infinite time and app shutdown doesnt work well in pytest
                continue

            await self.__camera.trigger_hires_capture(job_uuid)

    async def run(self):
        await self.setup()
        await asyncio.gather(self.job_task())



from pydantic_settings import SettingsConfigDict

from .base import CfgBaseSettings


class CfgTriggerPynng(CfgBaseSettings):
    model_config = SettingsConfigDict(env_prefix="trigger_")

    # server: str = Field(default="0.0.0.0")



import struct
import uuid
from dataclasses import dataclass


@dataclass
class ImageMessage:
    device_id: int
    jpg_bytes: bytes
    job_id: uuid.UUID | None = None

    _header_fmt = "iI16s"  # device_id, jpg_len, uuid (16 Bytes)

    def to_bytes(self) -> bytes:
        sid_bytes = self.job_id.bytes if self.job_id else b"\x00" * 16
        header = struct.pack(self._header_fmt, self.device_id, len(self.jpg_bytes), sid_bytes)
        return header + self.jpg_bytes

    @classmethod
    def from_bytes(cls, data: bytes) -> "ImageMessage":
        header_size = struct.calcsize(cls._header_fmt)
        device_id, jpg_len, uuid_bytes = struct.unpack(cls._header_fmt, data[:header_size])
        jpg_bytes = data[header_size : header_size + jpg_len]
        job_id = None if uuid_bytes == b"\x00" * 16 else uuid.UUID(bytes=uuid_bytes)
        return cls(device_id, jpg_bytes, job_id)



"""Main entry point for CameraApp with pluggable backends."""

import argparse
import asyncio
import importlib
import logging
import sys

from .app import CameraApp
from .backends.cameras.base import CameraBackend
from .backends.cameras.output.base import CameraOutput
from .backends.cameras.output.pynng import PynngCameraOutput
from .backends.triggers.input.pynng import PynngTriggerInput

logger = logging.getLogger(__name__)


# --- Registry ------------------------

CAMERA_CLASSES = ["Virtual", "Picam"]


# --- Backend Factory ---------------------------------------------------


def camera_factory(class_name: str, device_id: int, output_lores: CameraOutput, output_hires: CameraOutput) -> CameraBackend:
    module_path = f".backends.cameras.{class_name.lower()}"
    module = importlib.import_module(module_path, __package__)
    return getattr(module, class_name)(device_id, output_lores, output_hires)


def resolve_class_name(cli_value: str, registry: list[str]) -> str:
    """Map CLI lowercase value back to the canonical class name."""
    for cls in registry:
        if cli_value == cls.lower():
            return cls
    raise ValueError(f"Unknown backend: {cli_value}")


# --- Argparse ---------------------------------------------------


def parse_args(args):
    parser = argparse.ArgumentParser(description="CameraApp with pluggable backends")

    parser.add_argument(
        "--camera",
        choices=[c.lower() for c in CAMERA_CLASSES],
        default=CAMERA_CLASSES[0].lower(),
        help="Camera backend to use",
    )
    parser.add_argument(
        "--device-id",
        type=int,
        default=0,
        help="Device ID",
    )
    parser.add_argument(
        "--bind-ip",
        type=str,
        default="[::]",
        help="Bind pynng listener to interface. [::] binds to all IPv6 interfaces, which usually includes IPv4 also. Otherwise try 0.0.0.0.",
    )
    parser.add_argument(
        "--base-port",
        type=int,
        default=5550,
        help="Starting from base-port the app will listen. Use to start multiple instances on one host.",
    )

    return parser.parse_args(args)


# --- Main -------------------------------------------------------


def main(args=None, run_app: bool = True):
    fmt = "%(asctime)s [%(levelname)8s] %(message)s (%(filename)s:%(lineno)s)"
    logging.basicConfig(level=logging.DEBUG, format=fmt)

    args = parse_args(args)  # parse here, not above because pytest system exit 2

    port_input_trigger = args.base_port
    port_output_lores = args.base_port + 1
    port_output_hires = args.base_port + 2

    input_trigger = PynngTriggerInput(f"tcp://{args.bind_ip}:{port_input_trigger}")
    output_lores = PynngCameraOutput(f"tcp://{args.bind_ip}:{port_output_lores}")
    output_hires = PynngCameraOutput(f"tcp://{args.bind_ip}:{port_output_hires}")

    camera_class = resolve_class_name(args.camera, CAMERA_CLASSES)
    camera = camera_factory(camera_class, args.device_id, output_lores, output_hires)

    camera_app = CameraApp(camera, input_trigger)

    logger.info(f"Device Id: {args.device_id}")
    logger.info(f"Camera Backend: {camera_class}")
    logger.info(f"Service bound to {args.bind_ip} and ports [{port_input_trigger},{port_output_lores},{port_output_hires}]")

    try:
        if run_app:
            asyncio.run(camera_app.run())
    except KeyboardInterrupt:
        print("Exit app.")


if __name__ == "__main__":
    sys.exit(main(args=sys.argv[1:]))  # for testing



import pynng

from .base import CameraOutput


class PynngCameraOutput(CameraOutput):
    def __init__(self, address: str):
        self.__pub = pynng.Pub0()  # using pub instead push because we just want to broadcast and push would queue if not pulled
        self.__pub.listen(address)  # , block=False)
        # self.pub.listen("ipc:///home/michael/test.sock")

    def write(self, buf: bytes) -> int:
        """Synchronous send."""
        self.__pub.send(buf)

        return len(buf)

    async def awrite(self, buf: bytes) -> int:
        """Asynchronous send."""
        await self.__pub.asend(buf)
        return len(buf)



import asyncio
import io
import logging
import uuid

import numpy
from PIL import Image, ImageDraw

from ...config.camera_virtual import CfgCameraVirtual
from ...dto import ImageMessage
from .base import CameraBackend
from .output.base import CameraOutput

logger = logging.getLogger(__name__)


class Virtual(CameraBackend):
    """
    A fake camera backend that generates synthetic frames.
    Produces both 'lores' and 'hires' frames as byte strings.
    """

    def __init__(self, device_id: int, output_lores: CameraOutput, output_hires: CameraOutput):
        self.__config = CfgCameraVirtual()
        super().__init__(device_id, output_lores, output_hires)

        self.__offset_x = 0
        self.__offset_y = 0
        self.__color_current = 0

        logger.info(f"VirtualBackend initialized, {device_id=}, listening for subs")

    async def run(self):
        while True:
            # Offload CPU‑bound work to a thread
            produced_frame = await asyncio.to_thread(self._produce_dummy_image)

            # For demo, use same image for lores and hires
            msg_bytes = ImageMessage(self._device_id, jpg_bytes=produced_frame).to_bytes()
            await self._output_lores.awrite(msg_bytes)

            await asyncio.sleep(1.0 / self.__config.fps_nominal)

    async def trigger_hires_capture(self, job_id: uuid.UUID):
        produced_frame = await asyncio.to_thread(self._produce_dummy_image)

        msg_bytes = ImageMessage(self._device_id, jpg_bytes=produced_frame, job_id=job_id).to_bytes()
        await self._output_hires.awrite(msg_bytes)

    def _produce_dummy_image(self) -> bytes:
        """CPU-intensive image generator — run in a worker thread."""
        offset_x = self.__offset_x
        offset_y = self.__offset_y

        size = 250
        ellipse_divider = 3
        color_steps = 100
        byte_io = io.BytesIO()

        mask = Image.new("L", (size, size), 0)
        draw = ImageDraw.Draw(mask)
        draw.ellipse((0, 0, size // ellipse_divider, size // ellipse_divider), fill=255)

        time_normalized = self.__color_current / color_steps
        self.__color_current = self.__color_current + 1 if self.__color_current < color_steps else 0

        imarray = numpy.empty((size, size, 3))
        imarray[:, :, 0] = 0.5 + 0.5 * numpy.sin(2 * numpy.pi * (0 / 3 + time_normalized))
        imarray[:, :, 1] = 0.5 + 0.5 * numpy.sin(2 * numpy.pi * (1 / 3 + time_normalized))
        imarray[:, :, 2] = 0.5 + 0.5 * numpy.sin(2 * numpy.pi * (2 / 3 + time_normalized))
        imarray = numpy.round(255 * imarray).astype(numpy.uint8)

        random_image = Image.fromarray(imarray, "RGB")
        random_image.paste(mask, (size // ellipse_divider + offset_x, size // ellipse_divider + offset_y), mask=mask)

        random_image.save(byte_io, format="JPEG", quality=70)
        return byte_io.getvalue()



import asyncio
import io
import logging
import uuid

from libcamera import Transform, controls  # type: ignore
from picamera2 import Picamera2
from picamera2.devices.imx708 import IMX708
from picamera2.encoders import MJPEGEncoder, Quality
from picamera2.outputs.output import Output

from ...config.camera_picamera2 import CfgCameraPicamera2
from ...dto import ImageMessage
from .base import CameraBackend
from .output.base import CameraOutput

# Suppress debug logs from picamera2
logging.getLogger("picamera2").setLevel(logging.INFO)
logger = logging.getLogger(__name__)


class PicameraEncoderOutputAdapter(Output):
    def __init__(self, device_id: int, output: CameraOutput):
        self.__device_id = device_id
        self.__output = output

    def outputframe(self, frame, keyframe=True, timestamp=None, packet=None, audio=False):
        self.__output.write(ImageMessage(self.__device_id, jpg_bytes=frame, job_id=None).to_bytes())


class Picam(CameraBackend):
    def __init__(self, device_id: int, output_lores: CameraOutput, output_hires: CameraOutput):
        self.__config = CfgCameraPicamera2()
        super().__init__(device_id, output_lores, output_hires)

        self.__picamera2: Picamera2 | None = None
        self.__picamera2_output_lores = PicameraEncoderOutputAdapter(device_id, self._output_lores)

        logger.info(f"Picamera2Backend initialized, {device_id=}, listening for subs")

    async def trigger_hires_capture(self, job_id: uuid.UUID):
        jpeg_bytes = await asyncio.to_thread(self._produce_image)

        msg_bytes = ImageMessage(self._device_id, jpg_bytes=jpeg_bytes, job_id=job_id).to_bytes()
        await self._output_hires.awrite(msg_bytes)

    def _produce_image(self) -> bytes:
        assert self.__picamera2

        jpeg_buffer = io.BytesIO()
        self.__picamera2.capture_file(jpeg_buffer, format="jpeg")
        jpeg_bytes = jpeg_buffer.getvalue()

        return jpeg_bytes

    def _set_pi5_hdr(self, enable: bool):
        """enable/disable Pi5 specific HDR."""
        assert self.__picamera2
        if enable:
            self.__picamera2.set_controls({"HdrMode": controls.HdrModeEnum.SingleExposure})
        else:
            self.__picamera2.set_controls({"HdrMode": controls.HdrModeEnum.Off})

    def _set_imx708_hdr(self, enable: bool):
        """enable/disable imx708 (camera module 3) specific HDR. Resolution if enabled is max (Wxxxx,Hyyyy).
        Call before opening a Picamera2 object for regular use."""
        try:
            with IMX708(camera_num=self.__config.camera_num) as cam:
                if enable:
                    cam.set_sensor_hdr_mode(True)
                else:
                    cam.set_sensor_hdr_mode(False)
        except Exception as exc:
            logger.warning(f"could not set imx708 hdr mode, error: {exc}")

    async def run(self):
        # initialize private props

        logger.debug("starting _camera_fun")

        if self.__config.hdr_type == "imx708":
            self._set_imx708_hdr(True)

        self.__picamera2 = Picamera2(camera_num=self.__config.camera_num)

        if self.__config.hdr_type == "pi5":
            self._set_pi5_hdr(True)

        # configure; camera needs to be stopped before
        append_optmemory_format = {}
        if self.__config.optimize_memoryconsumption:
            logger.info("enabled memory optimization by choosing YUV420 format for main/lores streams")
            # if using YUV420 on main, also disable NoisReduction because it's done in software and causes framerate dropping on vc4 devices
            # https://github.com/raspberrypi/picamera2/discussions/1158#discussioncomment-11212355
            append_optmemory_format = {"format": "YUV420"}

        # configure; synchronization enabled?
        append_software_sync_control = {}
        if self.__config.software_sync == "server":
            logger.info("enabled synchronization, this node is configured as SERVER")
            append_software_sync_control = {"SyncMode": controls.rpi.SyncModeEnum.Server}
        elif self.__config.software_sync == "client":
            logger.info("enabled synchronization, this node is configured as CLIENT")
            append_software_sync_control = {"SyncMode": controls.rpi.SyncModeEnum.Client}
        else:
            logger.info("synchronization disabled.")

        camera_configuration = self.__picamera2.create_still_configuration(
            main={"size": (self.__config.camera_res_width, self.__config.camera_res_height), **append_optmemory_format},
            lores={"size": (self.__config.stream_res_width, self.__config.stream_res_height), **append_optmemory_format},
            encode="lores",
            display=None,
            buffer_count=3,  # 3 recommended if sync is used
            controls={
                "FrameRate": self.__config.framerate,
                # noise reduction might have impact on performance https://github.com/raspberrypi/picamera2/discussions/1158
                "NoiseReductionMode": controls.draft.NoiseReductionModeEnum.Minimal,
                **append_software_sync_control,
            },
            transform=Transform(hflip=self.__config.flip_horizontal, vflip=self.__config.flip_vertical),
        )
        self.__picamera2.configure(camera_configuration)

        self.__picamera2.start()

        try:
            self.__picamera2.set_controls({"AfMode": controls.AfModeEnum.Continuous})
        except RuntimeError as exc:
            logger.critical(f"control not available on camera - autofocus not working properly {exc}")

        try:
            self.__picamera2.set_controls({"AfSpeed": controls.AfSpeedEnum.Fast})
        except RuntimeError as exc:
            logger.info(f"control not available on all cameras - can ignore {exc}")

        logger.info(f"{self.__picamera2.camera_config=}")
        logger.info(f"{self.__picamera2.camera_controls=}")
        logger.info(f"{self.__picamera2.controls=}")
        logger.info(f"{self.__picamera2.camera_properties=}")

        mjpeg_encoder = MJPEGEncoder()
        mjpeg_encoder.frame_skip_count = self.__config.frame_skip_count
        self.__picamera2.start_recording(mjpeg_encoder, self.__picamera2_output_lores, quality=Quality[self.__config.videostream_quality])

        logger.debug(f"{self.__module__} started")

        if self.__config.software_sync != "off":
            logger.info("the node is configured to sync. ensure there is 1 server to sync to!")

        while True:
            # capture metadata blocks until new metadata is avail
            try:
                _ = await asyncio.to_thread(self.__picamera2.capture_metadata)

                # when sync client/server is enabled, the captures are synchronized by libcamera in the background
                # at one point there is the SyncTimer true. We do not supvervise it for now, so if there is no server
                # we don't know that the cameras are out of sync. Might improve later...
                # print("Sync ready:", meta.get("SyncReady"), "    Sync lag:", meta.get("SyncTimer"))

            except TimeoutError as exc:
                logger.warning(f"camera timed out: {exc}")
                break
